/* g3topbm.c - read a Group 3 FAX file and produce a portable bitmap
**
** Copyright (C) 1989 by Paul Haeberli <paul@manray.sgi.com>.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/

#include <stdio.h>
#include "pbm.h"
#include "g3.h"
#ifdef SYSV
#include <string.h>
#else /*SYSV*/
#ifdef VMS
#include <string.h>
#else
#include <strings.h>
#endif /* vax11c */
#endif /*SYSV*/

#define max(a,b) ((a) > (b) ? (a) : (b))

#define TABSIZE(tab) (sizeof(tab)/sizeof(struct tableentry))
#define MAXCOLS 1728
#define MAXROWS 4300	/* up to two pages long */

int eof = 0;
int eols;
int rawzeros;
int shdata;
int kludge;
int reversebits;
int stretch;

#define WHASHA 3510
#define WHASHB 1178

#define BHASHA 293
#define BHASHB 2695

#define HASHSIZE 1021
tableentry *whash[HASHSIZE];
tableentry *bhash[HASHSIZE];

char	*pm_progname;

main( argc, argv )
int argc;
char *argv[];
    {
    FILE *ifd;
#ifdef vax11c
    FILE *ofd;
#endif /* vax11c */
    int argn, rows, cols, row, col, i;
    bit *bits[MAXROWS];
    char *usage = "[-kludge][-reversebits][-stretch] [g3file]";
    char *p, out_filename[1024];
#ifdef vax11c
    {
       char *ep;

       pm_progname = strrchr (argv[0],']');
       pm_progname++;
       ep = strchr(pm_progname, '.');
       if (ep != NULL) *ep = '\0';
    }
#else
    pm_progname = argv[0];
#endif /* vax11c */

    argn = 1;
    kludge = 0;
    reversebits = 0;
    stretch = 0;

    /* Check for flags. */
    while ( argn < argc  && argv[argn][0] == '-' )
	{
	if ( strncmp(argv[argn],"-kludge",max(strlen(argv[argn]),2)) == 0 )
	    kludge = 1;
	else if ( strncmp(argv[argn],"-reversebits",max(strlen(argv[argn]),2)) == 0 )
	    reversebits = 1;
	else if ( strncmp(argv[argn],"-stretch",max(strlen(argv[argn]),2)) == 0 )
	    stretch = 1;
	else {
	    printf(usage); exit(1);
	}
	argn++;
	}

    if ( argn < argc )
	{
	ifd = fopen( argv[argn], "r" );
	argn++;
	}
    else
	ifd = stdin;

    if ( argn != argc ) {
	printf(usage); exit(1);
    }

    rows = MAXROWS;
    cols = MAXCOLS;
    eols = 0;

    if ( kludge )
	{
	/* Skip extra lines to get in sync. */
	skiptoeol( ifd );
	skiptoeol( ifd );
	skiptoeol( ifd );
	}
    skiptoeol( ifd );
    for ( i = 0; i < HASHSIZE; ++i )
	whash[i] = bhash[i] = (tableentry *) 0;
    addtohash( whash, twtable, TABSIZE(twtable), WHASHA, WHASHB );
    addtohash( whash, mwtable, TABSIZE(mwtable), WHASHA, WHASHB );
    addtohash( whash, extable, TABSIZE(extable), WHASHA, WHASHB );
    addtohash( bhash, tbtable, TABSIZE(tbtable), BHASHA, BHASHB );
    addtohash( bhash, mbtable, TABSIZE(mbtable), BHASHA, BHASHB );
    addtohash( bhash, extable, TABSIZE(extable), BHASHA, BHASHB );

    cols = 0;
    for ( rows = 0; rows < MAXROWS; ++rows )
	{
	bits[rows] = malloc( MAXCOLS );
	col = getfaxrow( ifd, rows, bits[rows] );
	if ( eof )
	    break;
	if ( col > cols )
	    cols = col;
	if ( stretch )
	    {
	    bits[rows + 1] = bits[rows];
	    ++rows;
	    }
	}

    fclose( ifd );

    i = argn - 1;
    strcpy(out_filename, argv[i]);
    if((p = strrchr(out_filename, '.')) != NULL) strcpy(p, ".ps");
    if((ofd = fopen(out_filename, "w")) == NULL) {
		perror(out_filename); exit(1);
    }

	fprintf(ofd, "%%!PS-Adobe-2.0 EPSF-2.0\n");
	fprintf(ofd, "%%%%Title: 4_0.ps\n");
	fprintf(ofd, "%%%%Creator: G3toFAX.\n");
	fprintf(ofd, "%%%%BoundingBox: 1 8 619 773\n");
	fprintf(ofd, "%%%%Pages: 1\n");
	fprintf(ofd, "%%%%DocumentFonts:\n");
	fprintf(ofd, "%%%%EndComments\n");
	fprintf(ofd, "%%%%EndProlog\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%%%%Page: 1 1\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% remember original state\n");
	fprintf(ofd, "/origstate save def\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% build a temporary dictionary\n");
	fprintf(ofd, "20 dict begin\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% define string to hold a scanline's worth of data\n");
	fprintf(ofd, "/pix 54 string def\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% lower left corner\n");
	fprintf(ofd, "1 8 translate\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% size of image (on paper, in 1/72inch coords)\n");
	fprintf(ofd, "618 765 scale\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% dimensions of data\n");
	fprintf(ofd, "%d %d 1\n", cols, rows);
	fprintf(ofd, "\n");
	fprintf(ofd, "%% mapping matrix\n");
	fprintf(ofd, "[%d 0 0 -%d 0 %d]\n", cols, rows, rows);
	fprintf(ofd, "\n");
	fprintf(ofd, "{currentfile pix readhexstring pop}\n");
	fprintf(ofd, "image\n");

printf("Writing image\n");
printf("Read %d rows, %d cols (MaxCol=%d)\n", rows, cols, MAXCOLS);	/*~~~*/
	for ( row = 0; row < rows; ++row ) {
		bit *bp = bits[row];
		char	*p, outbuf[128];
		p = outbuf;
		for(i = 0; i < cols; i += 4) {
			int j, out;
			out = 0xf;
			for(j = 3; j >= 0; j--)
				if(*bp++) out &= ~(1 << j);
			*p++ = (out > 9 ? out + 'a' - 10 : out + '0');
			if((i > 0) && (i % 70 == 0)) {
				*p = '\0';
				fprintf(ofd, "%s\n", outbuf);
				p = outbuf;
			}
		}
		*p = '\0';
		fprintf(ofd, "%s\n", outbuf);
	}
printf("Done\n");
	fprintf(ofd, "showpage\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% stop using temporary dictionary\n");
	fprintf(ofd, "end\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%% restore original state\n");
	fprintf(ofd, "origstate restore\n");
	fprintf(ofd, "\n");
	fprintf(ofd, "%%%%Trailer\n");

     fclose(ofd);
    exit(0);
    }

addtohash(hash, te, n, a, b)
	tableentry *hash[];
	tableentry *te;
	int n, a, b;
{
	unsigned int pos;

	while (n--) {
		pos = ((te->length+a)*(te->code+b))%HASHSIZE;
		if (hash[pos] != 0)
			printf(
			    "internal error: addtohash fatal hash collision",
			    0,0,0,0,0 );
		hash[pos] = te;
		te++;
	}
}

tableentry *
hashfind(hash, length, code, a, b)
	tableentry *hash[];
	int length, code;
	int a, b;
{
	unsigned int pos;
	tableentry *te;

	pos = ((length+a)*(code+b))%HASHSIZE;
	if (pos < 0 || pos >= HASHSIZE)
		printf(
		  "internal error: bad hash position, length %d code %d pos %d",
		    length, code, pos, 0,0 );
	te = hash[pos];
	return ((te && te->length == length && te->code == code) ? te : 0);
}

getfaxrow( inf, row, bitrow )
    FILE *inf;
    int row;
    bit *bitrow;
{
	int col;
	bit *bP;
	int curlen, curcode, nextbit;
	int count, color;
	tableentry *te;

	for ( col = 0, bP = bitrow; col < MAXCOLS; ++col, ++bP )
	    *bP = PBM_WHITE;
	col = 0;
	rawzeros = 0;
	curlen = 0;
	curcode = 0;
	color = 1;
	count = 0;
	while (!eof) {
		if (col >= MAXCOLS) {
			skiptoeol(inf);
			return (col); 
		}
		do {
			if (rawzeros >= 11) {
				nextbit = rawgetbit(inf);
				if (nextbit) {
					if (col == 0)
						/* XXX should be 6 */
						eof = (++eols == 3);
					else
						eols = 0;
#ifdef notdef
					if (col && col < 1728)
						printf(
					       "warning, row %d short (len %d)",
						    row, col, 0,0,0 );
#endif /*notdef*/
					return (col); 
				}
			} else
				nextbit = rawgetbit(inf);
			curcode = (curcode<<1) + nextbit;
			curlen++;
		} while (curcode <= 0);
		if (curlen > 13) {
			printf(
	  "bad code word at row %d, col %d (len %d code 0x%x), skipping to EOL",
			    row, col, curlen, curcode, 0 );
			skiptoeol(inf);
			return (col);
		}
		if (color) {
			if (curlen < 4)
				continue;
			te = hashfind(whash, curlen, curcode, WHASHA, WHASHB);
		} else {
			if (curlen < 2)
				continue;
			te = hashfind(bhash, curlen, curcode, BHASHA, BHASHB);
		}
		if (!te)
			continue;
		switch (te->tabid) {
		case TWTABLE:
		case TBTABLE:
			count += te->count;
			if (col+count > MAXCOLS) 
				count = MAXCOLS-col;
			if (count > 0) {
				if (color) {
					col += count;
					count = 0;
				} else {
					for ( ; count > 0; --count, ++col )
						bitrow[col] = PBM_BLACK;
				}
			}
			curcode = 0;
			curlen = 0;
			color = !color;
			break;
		case MWTABLE:
		case MBTABLE:
			count += te->count;
			curcode = 0;
			curlen = 0;
			break;
		case EXTABLE:
			count += te->count;
			curcode = 0;
			curlen = 0;
			break;
		default:
			printf( "internal bad poop", 0,0,0,0,0 );
		}
	}
	return (0);
}

skiptoeol(file)
FILE *file;
{
    int b;

    while (rawzeros<11)
	b = rawgetbit(file);
    while(1) {
	if(rawgetbit(file))
	    break;
    }
}

int shbit = 0;

int rawgetbit( file )
    FILE *file;
    {
    int b;

    if ( ( shbit & 0xff ) == 0 )
	{
	shdata = getc( file );
	if ( shdata == EOF ) {
/*	    printf( "premature EOF at line %d", eols ); */
		printf("Premature EOF...\n");	/*~~~*/
		eof = 1;
        }
	shbit = reversebits ? 0x01 : 0x80;
	}
    if ( shdata & shbit )
	{
	rawzeros = 0;
	b = 1;
	}
    else
	{
	rawzeros++;
	b = 0;
	}
    if ( reversebits )
	shbit <<= 1;
    else
	shbit >>= 1;
    return b;
    }
