/*
  This file is part of the NetFax system.

  (c) Copyright 1989 by David M. Siegel and Sundar Narasimhan.
      All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
VMS notes: We use VMS $QIO in order to not interfere with Multinet's write.
*/

#include <stdio.h>
#ifdef VMS
#include <file.h>
#include "../../include/vms.h"
#else /* VMS */
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#endif /* VMS */

#include "log.h"
#include "tty.h"

/*
 | On VMS we assume hardware flow control thus all the mungling here is
 | ignored.
 */
int tty_fc(fd, state)
     int fd;
     fc_state state;
{
#ifndef VMS
    struct termios t;
    
    if (tcgetattr(fd, &t) < 0) {
	log(L_EMERG, "tty_fc: can't get attrs: %m");
	return (-1);
    }

    switch (state) {
      case FC_OUTPUT_ON:
	t.c_iflag |= IXON;
	t.c_iflag &= ~IXOFF;
	break;
      case FC_INPUT_ON:
	t.c_iflag |= IXOFF;
	t.c_iflag &= ~IXON;
	break;
      case FC_BOTH_ON:
	t.c_iflag |= (IXON|IXOFF);
	break;
      case FC_BOTH_OFF:
	t.c_iflag &= ~(IXON|IXOFF);
	break;
    }

    if (tcsetattr(fd, TCSANOW, &t) < 0) {
	log(L_EMERG, "tty_fc: can't set attrs: %m");
	return (-1);
    }
#endif	/* VMS */

    return (0);
}    

/*
 | On VMS the modem port will be set in advance to be:
 | /PASTHRU/EIGHT/COMMSYNC/SPEED=19200  (/NOMODEM is needed for CommSync).
 */
static int condition_tty(fd)
     int fd;
{
#ifndef VMS
    struct termios t;

    if (tcgetattr(fd, &t) < 0) {
	log(L_EMERG, "condition_tty: can't get attrs: %m");
	return (-1);
    }

    t.c_iflag = IXON|IXOFF;
    t.c_oflag = 0;
    t.c_cflag = CS8|CREAD|CLOCAL;
    t.c_lflag = 0;


    cfsetispeed(&t, B19200);
    cfsetospeed(&t, B19200);

    if (tcsetattr(fd, TCSANOW, &t) < 0) {
	log(L_EMERG, "condition_tty: can't set attrs: %m");
	return (-1);
    }
#endif /* VMS */

    return (0);
}

#ifdef VMS
int tty_open(filename)
     char *filename;
{
	struct DESC	FaxDevice = { strlen(filename), 0, filename};
	int	fd, status;

	if(((status = sys$assign(&FaxDevice, &fd, 0, 0)) & 0x1) == 0) {
		printf("Can't assign channel to '%s', status=%d\n",
			filename, status);
		log(L_EMERG, "can't open fax modem file: %m");
		return (-1);
	}
	return(fd);
}

int tty_close(fd)
     int fd;
{
	return(sys$dassgn(fd));
}
#else	/* VMS */
int tty_open(filename)
     char *filename;
{
    int fd;

    if ((fd = open(filename, O_RDWR|O_NDELAY)) < 0) {
	log(L_EMERG, "can't open fax modem file: %m");
	return (-1);
    }
    
    if (condition_tty(fd) < 0) {
	close(fd);
	return (-1);
    }

    return (fd);
}

int tty_close(fd)
     int fd;
{
    return (close(fd));
}
#endif /* VMS */
